@@ -8,7 +8,10 @@ use smithay_client_toolkit::reexports::client;
 
 use client::protocol::{wl_compositor, wl_seat, wl_shm, wl_subcompositor, wl_surface};
 use client::{Attached, DispatchData};
-use tiny_skia::{FillRule, Paint, Path, PathBuilder, Pixmap, Point, Rect, Transform};
+use tiny_skia::{
+    ClipMask, Color, FillRule, Paint, Path, PathBuilder, Pixmap, PixmapMut, PixmapPaint, Point,
+    Rect, Transform,
+};
 
 use smithay_client_toolkit::{
     seat::pointer::{ThemeManager, ThemeSpec, ThemedPointer},
@@ -21,16 +24,14 @@ use theme::{ColorTheme, BORDER_SIZE, HEADER_SIZE};
 
 mod buttons;
 use buttons::{ButtonKind, Buttons};
+use title::TitleText;
 
 use crate::theme::ColorMap;
 
-mod font;
-mod minitype;
 mod parts;
 mod pointer;
 mod surface;
-
-const ROBOTO_FONT: &[u8] = include_bytes!("../assets/Roboto-Regular.ttf");
+mod title;
 
 /*
  * Utilities
@@ -158,8 +159,8 @@ pub struct AdwaitaFrame {
 
     buttons: Rc<RefCell<Buttons>>,
     colors: ColorTheme,
-    font: font::Font,
     title: Option<String>,
+    title_text: TitleText,
 }
 
 impl Frame for AdwaitaFrame {
@@ -194,6 +195,8 @@ impl Frame for AdwaitaFrame {
 
         let pool = AutoMemPool::new(shm.clone())?;
 
+        let colors = ColorTheme::default();
+
         Ok(AdwaitaFrame {
             base_surface: base_surface.clone(),
             compositor: compositor.clone(),
@@ -206,9 +209,9 @@ impl Frame for AdwaitaFrame {
             themer,
             surface_version: compositor.as_ref().version(),
             buttons: Default::default(),
-            colors: Default::default(),
-            font: font::Font::from_bytes(ROBOTO_FONT).unwrap(),
             title: None,
+            title_text: TitleText::new(colors.active.font_color).unwrap(),
+            colors,
         })
     }
 
@@ -336,6 +339,9 @@ impl Frame for AdwaitaFrame {
                 } else {
                     &self.colors.inactive
                 };
+
+                self.title_text.update_color(colors.font_color);
+
                 let border_paint = colors.border_paint();
 
                 // -> head-subsurface
@@ -345,16 +351,19 @@ impl Frame for AdwaitaFrame {
                     4 * header_width as i32,
                     wl_shm::Format::Argb8888,
                 ) {
-                    let mut pixmap = Pixmap::new(header_width, header_height).unwrap();
+                    let mut pixmap =
+                        PixmapMut::from_bytes(canvas, header_width, header_height).unwrap();
+                    pixmap.fill(Color::TRANSPARENT);
+
+                    self.title_text.update_scale(header_scale);
 
                     draw_headerbar(
                         &mut pixmap,
+                        self.title_text.pixmap(),
                         header_scale as f32,
                         inner.resizable,
                         self.active,
                         &self.colors,
-                        &self.font,
-                        self.title.as_ref(),
                         &self.buttons.borrow(),
                         &self
                             .pointers
@@ -371,16 +380,6 @@ impl Frame for AdwaitaFrame {
                             .collect::<Vec<Location>>(),
                     );
 
-                    let buff = pixmap.data();
-
-                    for (id, pixel) in canvas.iter_mut().enumerate() {
-                        *pixel = buff[id];
-                    }
-
-                    // decoration
-                    //     .header
-                    //     .subsurface
-                    //     .set_position(-(BORDER_SIZE as i32), -(HEADER_SIZE as i32));
                     decoration.header.subsurface.set_position(
                         -(BORDER_SIZE as i32),
                         -(HEADER_SIZE as i32 + BORDER_SIZE as i32),
@@ -419,7 +418,8 @@ impl Frame for AdwaitaFrame {
                     (4 * bottom_scale * (width + 2 * BORDER_SIZE)) as i32,
                     wl_shm::Format::Argb8888,
                 ) {
-                    let mut pixmap = Pixmap::new(w as u32, h as u32).unwrap();
+                    let mut pixmap = PixmapMut::from_bytes(canvas, w as u32, h as u32).unwrap();
+                    pixmap.fill(Color::TRANSPARENT);
 
                     let size = 1.0;
                     let x = BORDER_SIZE as f32 * bottom_scale as f32 - 1.0;
@@ -436,11 +436,6 @@ impl Frame for AdwaitaFrame {
                         None,
                     );
 
-                    let data = pixmap.data();
-                    for (id, pixel) in canvas.iter_mut().enumerate() {
-                        *pixel = data[id];
-                    }
-
                     decoration
                         .bottom
                         .subsurface
@@ -478,7 +473,8 @@ impl Frame for AdwaitaFrame {
                     let mut bg = Paint::default();
                     bg.set_color_rgba8(255, 0, 0, 255);
 
-                    let mut pixmap = Pixmap::new(w as u32, h as u32).unwrap();
+                    let mut pixmap = PixmapMut::from_bytes(canvas, w as u32, h as u32).unwrap();
+                    pixmap.fill(Color::TRANSPARENT);
 
                     let size = 1.0;
                     pixmap.fill_rect(
@@ -488,11 +484,6 @@ impl Frame for AdwaitaFrame {
                         None,
                     );
 
-                    let data = pixmap.data();
-                    for (id, pixel) in canvas.iter_mut().enumerate() {
-                        *pixel = data[id];
-                    }
-
                     decoration
                         .left
                         .subsurface
@@ -525,7 +516,8 @@ impl Frame for AdwaitaFrame {
                     let mut bg = Paint::default();
                     bg.set_color_rgba8(255, 0, 0, 255);
 
-                    let mut pixmap = Pixmap::new(w as u32, h as u32).unwrap();
+                    let mut pixmap = PixmapMut::from_bytes(canvas, w as u32, h as u32).unwrap();
+                    pixmap.fill(Color::TRANSPARENT);
 
                     let size = 1.0;
                     pixmap.fill_rect(
@@ -535,11 +527,6 @@ impl Frame for AdwaitaFrame {
                         None,
                     );
 
-                    let data = pixmap.data();
-                    for (id, pixel) in canvas.iter_mut().enumerate() {
-                        *pixel = data[id];
-                    }
-
                     decoration.right.subsurface.set_position(width as i32, 0);
                     decoration.right.surface.attach(Some(&buffer), 0, 0);
                     if self.surface_version >= 4 {
@@ -587,6 +574,7 @@ impl Frame for AdwaitaFrame {
     }
 
     fn set_title(&mut self, title: String) {
+        self.title_text.update_title(&title);
         self.title = Some(title);
     }
 }
@@ -602,13 +590,12 @@ impl Drop for AdwaitaFrame {
 }
 
 fn draw_headerbar(
-    pixmap: &mut Pixmap,
+    pixmap: &mut PixmapMut,
+    text_pixmap: Option<&Pixmap>,
     scale: f32,
     maximizable: bool,
     state: WindowState,
     colors: &ColorTheme,
-    font: &font::Font,
-    title: Option<&String>,
     buttons: &Buttons,
     mouses: &[Location],
 ) {
@@ -621,18 +608,15 @@ fn draw_headerbar(
 
     draw_headerbar_bg(pixmap, scale, margin_h, margin_v, colors);
 
-    if let Some(title) = title {
-        let size = 20.0 * scale as f64;
-
+    if let Some(text_pixmap) = text_pixmap {
         let canvas_w = pixmap.width() as f32;
         let canvas_h = pixmap.height() as f32;
 
         let header_w = canvas_w - margin_h * 2.0;
         let header_h = canvas_h - margin_v;
 
-        let (w, h) = font.measure_text(title, size);
-        let text_w = w as f32;
-        let text_h = h as f32;
+        let text_w = text_pixmap.width() as f32;
+        let text_h = text_pixmap.height() as f32;
 
         let x = header_w / 2.0 - text_w / 2.0;
         let y = header_h / 2.0 - text_h / 2.0;
@@ -652,14 +636,24 @@ fn draw_headerbar(
 
         let x = x.max(margin_h + 5.0);
 
-        font.render_text(
-            size,
-            &colors.font_paint(),
-            pixmap,
-            (x as f64, y as f64),
-            title,
-            buttons.minimize.x() as f64 - 10.0,
-        );
+        if let Some(clip) = Rect::from_xywh(0.0, 0.0, buttons.minimize.x() - 10.0, header_h) {
+            let mut mask = ClipMask::new();
+            mask.set_path(
+                canvas_w as u32,
+                canvas_h as u32,
+                &PathBuilder::from_rect(clip),
+                FillRule::Winding,
+                false,
+            );
+            pixmap.draw_pixmap(
+                x as i32,
+                y as i32,
+                text_pixmap.as_ref(),
+                &PixmapPaint::default(),
+                Transform::identity(),
+                Some(&mask),
+            );
+        }
     }
 
     if buttons.close.x() > margin_h {
@@ -680,7 +674,7 @@ fn draw_headerbar(
 }
 
 fn draw_headerbar_bg(
-    pixmap: &mut Pixmap,
+    pixmap: &mut PixmapMut,
     scale: f32,
     margin_h: f32,
     margin_v: f32,
